/*

  RepRap
  ------

  The Replicating Rapid Prototyper Project


  Copyright (C) 2005
  Adrian Bowyer & The RepRap Researchers

  http://reprap.org

  Principal author:

     Adrian Bowyer
     Department of Mechanical Engineering
     Faculty of Engineering and Design
     University of Bath
     Bath BA2 7AY
     U.K.

     e-mail: A.Bowyer@bath.ac.uk

  RepRap is free; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  Licence as published by the Free Software Foundation; either
  version 2 of the Licence, or (at your option) any later version.

  RepRap is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public Licence for more details.

  For this purpose the words "software" and "library" in the GNU Library
  General Public Licence are taken to mean any and all computer programs
  computer files data results documents and other copyright information
  available from the RepRap project.

  You should have received a copy of the GNU Library General Public
  Licence along with RepRap; if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA,
  or see

      http://www.gnu.org/

==================================================================================

This is the C program for the PIC that controls magnetically encoded servo.

18 May 2006

*/

#include "servo.h"

// PIC configuration bits

typedef unsigned int config;
config at 0x2007 __CONFIG = _CP_OFF &
  _WDT_OFF &
  _BODEN_OFF &
  _PWRTE_ON &
  _INTRC_OSC_CLKOUT &
  _MCLRE_OFF &
  _LVP_OFF;


// Globals

long count, target, temp, diff, brake, time;
byte old_b, c_in, max_speed, speed, t0, t1, t2, t3, forward;

static void isr() interrupt 0 
{
  byte b1, b2;

  if(RBIF)
    {
      b1 = PORTB & 0xc0;
      b2 = b1 | old_b;

      switch(b2)
	{
	case 0x10:
	case 0x70:
	case 0x80:
	case 0xe0:
	  count++;
	  forward = 1;
	  break;

	case 0x20:
	case 0x40:
	case 0xb0:
	case 0xd0:
	  count--;
	  forward = 0;
	  break;

	default:
	  sendByte('!');
	}

      RBIF = 0;
      old_b = b1 >> 2;
    }

  if(RCIF)
    {
      c_in = RCREG;
      RCIF = 0;
      RCREG = 0;
    }
}

// ************************************************************************

// long <-> bytes

void toBytes()
{
  t0 = temp & 0xff;
  temp = temp >> 8;
  t1 = temp & 0xff;
  temp = temp >> 8;
  t2 = temp & 0xff;
  temp = temp >> 8;
  t3 = temp & 0xff;
}

void toLong()
{
  temp = t3;
  temp = temp << 8;
  temp = temp | t2;
  temp = temp << 8;
  temp = temp | t1;
  temp = temp << 8;
  temp = temp | t0;
}

// **********************************************************************

// Motor control

void motorForward()
{
    CCPR1L = speed;
    CCP1CON = BIN(00001100);
    PORTB4 = 1;
    PORTB5 = 0;
}

void motorReverse()
{
    CCPR1L = speed;
    CCP1CON = BIN(00001100);
    PORTB4 = 0;
    PORTB5 = 1;
}

void motorHalt()
{
  CCP1CON = 0;
  PORTB4 = 0;
  PORTB5 = 0;
}

byte close()
{
  if(diff > -2 && diff < 2)
    {
      motorHalt();
      return 1;
    }
  return 0;
}

void feedback()
{
  byte s_speed;

  diff = target - count;

  if(close()) return;

  if(diff > -brake && diff < brake)
  {
    if(forward)
      motorReverse();
    else
      motorForward();
    return;
  }

  if(diff < 0)
    motorReverse();
  else if(diff > 0)
    motorForward();
  else
    motorHalt();
}

void setSpeed(byte s)
{
  speed = s;
  if(speed > max_speed)
    speed = max_speed;
  CCPR1L = speed;
}


// ***********************************************************************

// COMMUNICATIONS


// Send a byte out of the PIC's USART.

void sendByte(byte c)
{
  while(!TRMT); 
  TXREG = c;
}

// Send a byte in hex.

void sendNumber(byte n)
{
  byte b;

  b = n >> 4;
  if(b >= 10)
    sendByte(b + ('a' - 10));
  else
    sendByte(b + '0');
  b = n & 0x0f;
  if(b >= 10)
    sendByte(b + ('a' - 10));
  else
    sendByte(b + '0');
}


// Send temp in hex.

void sendTemp()
{
  toBytes();
  sendNumber(t3);
  sendNumber(t2);
  sendNumber(t1);
  sendNumber(t0);
}

// Wait for keyboard input (N.B. feedback() will not be being called when this happens)

byte waitByte()
{
  byte b;
  while(!c_in);
  b = c_in;
  c_in = 0;
  sendByte(b); // Echo
  return b;
}

// Get a number in hex.

byte waitNumber()
{
  byte b, n;
  b = waitByte();
  if(b > '9')
    b = b - 'a' + 10;
  else
    b = b - '0';
  n = waitByte();
  if(n > '9')
    n = n - 'a' + 10;
  else
    n = n - '0';
  n = n | (b << 4);
  return n;
}

void waitTemp()
{
  t3 = waitNumber();
  t2 = waitNumber();
  t1 = waitNumber();
  t0 = waitNumber();
  toLong();
}

// Send a new line.

void crlf()
{
  sendByte('\r');
  sendByte('\n');
}

// Send a "> " prompt.

void prompt()
{
  crlf();
  sendByte('>');
  sendByte(' ');
}

// Deal with user input (single letter command)

void menu()
{
  byte b, c;
  
  b = c_in;
  c_in = 0;

  sendByte(b); // Echo

  switch(b)
    {

    // return and set the number of brake steps

    case 'b':
      c = brake & 0xff;
      sendNumber(c);
      break;

    case 'B':
      brake = waitNumber();
      break;

    // Return the current step count

    case 'c':
      temp = count;
      sendTemp();
      break;

    // Zero

    case '0':
      count = 0;
      target = 0;
      break;

    // Halt any movement by setting the target to the current position

    case 'H':
      target = count;
      break;

    // Set and return the maximum speed

    case 'S':
      setSpeed(waitNumber());
      break;

    case 's':
      sendNumber(speed);
      break;

    // Set and return the maximum speed

    case 'M':
      max_speed = waitNumber();
      if(speed > max_speed)
	setSpeed(max_speed);
      break;

    case 'm':
      sendNumber(max_speed);
      break;

    // Return and set the target position

    case 't':
      temp = target;
      sendTemp();
      break;

    case 'T':
      waitTemp();
      target = temp;
      break;
      
    // Uh?

    default:
      sendByte('?');
    }
  prompt();
}


// ***************************************************************************

// Set everything up inside the PIC (init1) and the algorithm (init2)

void init1()
{
  OPTION_REG = BIN(01011111); // Disable TMR0 on RA4, 1:128 WDT, pullups on
  CMCON = 0xff;               // Comparator module defaults
  TRISA = BIN(00101100);      // Port A outputs except 2,3 (sync)
                              // RA3 is input for max limit sensor
                              // RA5 can only be used as an input
  TRISB = BIN(11000111);      // Port B 0-2 input, 3-7 outputs
                              // Port B3 is PWM output
  PIE1 = BIN(00000000);       // All peripheral interrupts initially disabled
  INTCON = BIN(00000000);     // Interrupts disabled
  PIR1 = 0;                   // Clear peripheral interrupt flags
  SPBRG = 12;                 // 12 = 19200 (actually 19230) baud @ 4MHz
  TXSTA = BIN(00000100);      // 8 bit high speed 
  RCSTA = BIN(10000000);      // Enable port for 8 bit receive

  PORTB = BIN(00000001);      // Turn on pullups for B0,3

  RCIE = 1;  // Enable receive interrupts
  CREN = 1;  // Start reception

  TXEN = 1;  // Enable transmit
  RBIE = 1;  // Interrupt on B change
  PEIE = 1;  // Peripheral interrupts on
  GIE = 1;   // Now turn on interrupts

  PORTB = 0;
  PORTA = 0;

  TMR1IE = 0;

  CCPR1L = 0;
  PR2 = 0xff;
  T2CON = BIN(00000111);      // Start timer 1:x prescale, 1:x postscale
  CCP1CON = 0;

  TMR2 = 0;

  T1CON = BIN(00000000);  // Timer 1 in clock mode with 1:1 scale
  TMR1IE = 1;  // Enable timer interrupt
}

void init2()
{
  count = 0;
  target = 0;
  diff = 0;
  time = 0;
  brake = 7;
  old_b = 0;
  c_in = 0;
  max_speed = 64;
  speed = max_speed;
  forward = 0;
  motorHalt();
}


// Main program initialises everything, then runs an infinite loop listening
// for commands and controlling the peripherals by repeated function calls -
// i.e. crude process timesharing without interrupts.  Note menu items
// that need further input stop the loop (but not interrupts) until that 
// input is complete.

void main()
{
  init1();
  init2();
  prompt();

  while(1)
    {
      feedback();
      if(c_in) menu();
    }
}

